/* Sample scilab interface functio for getting functions, gradients, and dense Hessians from an AMPL .nl file.
   Start with:

	[asl,x,bl,bu,v,cl,cu] = ampl_init(<filename>)

   to read in a problem (discarding the previous problem, if any).
   The return values are:

	x = primal initial guess
	v = dual initial guess
	bl, bu = lower and upper bounds on x
	cl, cu = lower and upper bounds on c (the constraint bodies).

   Then

	[f,c] = ampl_evalf(asl,x)

   gives the function (f) and constraint bodies (c) at x;

	[g,Jac] = ampl_evalg(asl,x)

   gives the gradient g of f, the Jacobian matrix J of c, and

	W = ampl_evalw(asl,v)

   gives the Hessian W of the Lagrangian function L = f + v*c
   (at the last x at which amplfunc(x,1) was called).

   After finding optimal values for x and v,

	ampl_write_sol(asl,'solution message',x,v)

   to write a stub.sol file.
*/

#ifdef WIN32
#include <windows.h>
#endif

#include <string.h>

#include <stack-c.h>
#include <sciprint.h>
#include <MALLOC.h>

#include <asl_pfgh.h>

//#define DEBUG 1
#define DBGPRINTF sciprint
//#define DBGPRINTF printf

#ifdef _WIN32
/* Omit sw "signal" catching and x86 precision adjustment. */
#define ASL_NO_FP_INIT
#include <fpinit.c>
#endif /* _WIN32 */

#undef CreateVar
//#define CreateVar(n,ct,mx,nx,lx) if(! C2F(createvar)((c_local=n,&c_local),ct,mx,nx,lx, 1L)) { return 0;  }
#define CreateVar(n,ct,mx,nx,lx) C2F(createvar)((c_local=n,&c_local),ct,mx,nx,lx, 1L)

///////////////////////////////////////////////////////////////
// Load and get the initial values of the given AMPL problem //
///////////////////////////////////////////////////////////////

struct my_asl
{
  ASL * asl_pointer;
  real * J;
};

int sci_ampl_init(char * fname)
{
  int m_pointer = 1, n_pointer = 1;
  int m_string, n_string, l_string;
  int m_X0,     n_X0,     l_X0;
  int m_LUv,    n_LUv,    l_LUv;
  int m_UVx,    n_Uvx,    l_Uvx;
  int m_pi0,    n_pi0,    l_pi0;
  int m_LUrhs,  n_LUrhs,  l_LUrhs;
  int m_Urhsx,  n_Urhsx,  l_Urhsx;
  int n, nc, nz;
  struct my_asl * my_asl_pointer = NULL;
  FILE * nl;

  // The first parameter is a string, then, we open the given file
  if (((Rhs!=1)&&(GetType(1)!=sci_strings))||(Lhs!=7))
    {
      sciprint("ampl_init: [asl,x,bl,bu,v,cl,cu] = ampl_init(filename)\n");
      return 0;
    }

  GetRhsVar(1, "c", &m_string, &n_string, &l_string);

  my_asl_pointer = (struct my_asl *)MALLOC(1*sizeof(struct my_asl));

  my_asl_pointer->asl_pointer = ASL_alloc(ASL_read_pfgh);

  /* 0 ==> jacdim0 should exit if filename.nl does not exist; 
     1 ==> return 0 */
  my_asl_pointer->asl_pointer->i.return_nofile_ = 1;
  
  if (!(nl = jac0dim_ASL(my_asl_pointer->asl_pointer, cstk(l_string),strlen(cstk(l_string))))) 
    {
      Scierror(999,"ampl_init: can't open %s\n", cstk(l_string));
      ASL_free(&(my_asl_pointer->asl_pointer));
      return 0;
    }

  if (my_asl_pointer->asl_pointer->i.n_obj_ <= 0) sciprint("ampl_init: warning: objective == 0\n");
  
  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;
  nz = my_asl_pointer->asl_pointer->i.nzc_;

  my_asl_pointer->J = (real *)M1alloc_ASL(&(my_asl_pointer->asl_pointer->i),nz*sizeof(real));

  //  nhnz = sphsetup(0, 0, nc > 0, 0); // For the sparse things.

  m_X0    = n;  n_X0    = 1;
  m_LUv   = n;  n_LUv   = 1;
  m_UVx   = n;  n_Uvx   = 1;
  m_pi0   = nc; n_pi0   = 1;
  m_LUrhs = nc; n_LUrhs = 1;
  m_Urhsx = nc; n_Urhsx = 1;
  
  CreateVarFromPtr(Rhs+1, "p", &m_pointer, &n_pointer, (void*)my_asl_pointer);
  CreateVar(Rhs+2, "d", &m_X0,    &n_X0,    &l_X0);
  CreateVar(Rhs+3, "d", &m_LUv,   &n_LUv,   &l_LUv);
  CreateVar(Rhs+4, "d", &m_UVx,   &n_Uvx,   &l_Uvx);
  CreateVar(Rhs+5, "d", &m_pi0,   &n_pi0,   &l_pi0);
  CreateVar(Rhs+6, "d", &m_LUrhs, &n_LUrhs, &l_LUrhs);
  CreateVar(Rhs+7, "d", &m_Urhsx, &n_Urhsx, &l_Urhsx);
  
  my_asl_pointer->asl_pointer->i.X0_    = stk(l_X0);
  my_asl_pointer->asl_pointer->i.LUv_   = stk(l_LUv);
  my_asl_pointer->asl_pointer->i.Uvx_   = stk(l_Uvx);
  my_asl_pointer->asl_pointer->i.pi0_   = stk(l_pi0);
  my_asl_pointer->asl_pointer->i.LUrhs_ = stk(l_LUrhs);
  my_asl_pointer->asl_pointer->i.Urhsx_ = stk(l_Urhsx);

  pfgh_read_ASL(my_asl_pointer->asl_pointer, nl, ASL_findgroups);

  LhsVar(1) = Rhs + 1;
  LhsVar(2) = Rhs + 2;
  LhsVar(3) = Rhs + 3;
  LhsVar(4) = Rhs + 4;
  LhsVar(5) = Rhs + 5;
  LhsVar(6) = Rhs + 6;
  LhsVar(7) = Rhs + 7;

  return 0;
}

//////////////////////////////////////////////////////
// Release the memory allocated to the AMPL problem //
//////////////////////////////////////////////////////

int sci_ampl_free(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  struct my_asl * my_asl_pointer = NULL;

  if (Rhs!=1)
    {
      sciprint("ampl_free(asl)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  ASL_free(&(my_asl_pointer->asl_pointer));

  if (my_asl_pointer) FREE(my_asl_pointer);

  return 0;
}

///////////////////////////////////////////
// Computation of the objective function //
///////////////////////////////////////////

int sci_ampl_evalf(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_x,       n_x,       l_x;
  int m_f,       n_f,       l_f;
  int m_c,       n_c,       l_c;
  int n, nc, nerror;
  char * what;
  struct my_asl * my_asl_pointer = NULL;
  Jmp_buf err_jmp0;
  
  if (Rhs!=2) 
    {
      sciprint("[f,c] = ampl_evalf(asl, x)");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  nerror = -1;
  my_asl_pointer->asl_pointer->i.err_jmp1_ = &err_jmp0;

  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }
      
  CheckVector(2,n,1);
      
  GetRhsVar(2, "d", &m_x, &n_x, &l_x);

  m_f = 1;  n_f = 1;
  m_c = nc; n_c = 1;

  CreateVar(Rhs+1, "d", &m_f, &n_f, &l_f);
  CreateVar(Rhs+2, "d", &m_c, &n_c, &l_c);

  if (setjmp(err_jmp0.jb))
    {
      sciprint("ampl_evalf: trouble evaluating %s\n", what);
      return 0;
    }

  what = "f";
  *stk(l_f) = (*((ASL*)(my_asl_pointer->asl_pointer))->p.Objval)((ASL*)my_asl_pointer->asl_pointer,0,stk(l_x),&nerror);
#ifdef DEBUG
  DBGPRINTF("objval: nerror = %d\n", nerror);
#endif

  what = "c";
  (*((ASL*)(my_asl_pointer->asl_pointer))->p.Conval)((ASL*)(my_asl_pointer->asl_pointer),stk(l_x),stk(l_c),&nerror);
#ifdef DEBUG
  DBGPRINTF("conval: nerror = %d\n", nerror);
#endif
  
  LhsVar(1) = Rhs + 1;
  LhsVar(2) = Rhs + 2;

  return 0;
}

///////////////////////////////////
// Dense version of the Jacobian //
///////////////////////////////////

int sci_ampl_evalg(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_x,       n_x,       l_x;
  int m_g,       n_g,       l_g;
  int m_J1,      n_J1,      l_J1;
  int n, nc, nz, nerror;
  char   * what = NULL;
  double * tmp_dbl = NULL;
  real   * J = NULL;
  struct my_asl * my_asl_pointer = NULL;
  cgrad  * cg = NULL, ** cgp = NULL, ** cgpe = NULL;
  Jmp_buf err_jmp0;

  if (Rhs!=2) 
    {
      sciprint("[g,Jac] = ampl_evalg(asl,x)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }
      
  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;
  nz = my_asl_pointer->asl_pointer->i.nzc_;

  CheckVector(2,n,1);

  GetRhsVar(2, "d", &m_x, &n_x, &l_x);

  J = my_asl_pointer->J;

  nerror = -1;
  my_asl_pointer->asl_pointer->i.err_jmp1_ = &err_jmp0;

  m_g  = n;  n_g  = 1;
  m_J1 = nc; n_J1 = n;

  CreateVar(Rhs+1, "d", &m_g,  &n_g,  &l_g);
  CreateVar(Rhs+2, "d", &m_J1, &n_J1, &l_J1);

  if (setjmp(err_jmp0.jb))
    {
      sciprint("ampl_evalg: trouble evaluating %s\n", what);
      return 0;
    }

  what = "g";
  (*((ASL*)(my_asl_pointer->asl_pointer))->p.Objgrd)((ASL*)(my_asl_pointer->asl_pointer),0,stk(l_x),stk(l_g),&nerror);
  
  if (nc) 
    {
      what = "J";
      (*((ASL*)(my_asl_pointer->asl_pointer))->p.Jacval)((ASL*)(my_asl_pointer->asl_pointer),stk(l_x),J,&nerror);
      cgp = my_asl_pointer->asl_pointer->i.Cgrad_;

      memset(stk(l_J1),0,n*nc*sizeof(double));
      tmp_dbl = stk(l_J1);
      for(cgpe=cgp+nc; cgp<cgpe; tmp_dbl++)
	for(cg=*cgp++; cg; cg=cg->next)
	  tmp_dbl[nc*cg->varno] = J[cg->goff];
    }
  
  LhsVar(1) = Rhs + 1;
  LhsVar(2) = Rhs + 2;
  
  return 0;
}

//////////////////////////////////
// Dense version of the Hessian //
//////////////////////////////////

int sci_ampl_evalw(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_v,       n_v,       l_v;
  int m_W,       n_W,       l_W;
  int n, nc, nz;
  int nerror;
  char * what;
  struct my_asl * my_asl_pointer = NULL;
  Jmp_buf err_jmp0;
  
  if ((Lhs!=1)||(Rhs!=2))
    {
      sciprint("W = ampl_evalw(asl, v)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;
  nz = my_asl_pointer->asl_pointer->i.nzc_;

  CheckVector(2,nc,1);
  
  nerror = -1;
  my_asl_pointer->asl_pointer->i.err_jmp1_ = &err_jmp0;

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }

  GetRhsVar(2, "d", &m_v, &n_v, &l_v);
  
  m_W = n; n_W = n;
  CreateVar(Rhs+1, "d", &m_W, &n_W, &l_W);

  if (setjmp(err_jmp0.jb))
    {
      sciprint("ampl_evalw: trouble evaluating %s\n", what);
      return 0;
    }

  what = "W";
  (*((ASL*)(my_asl_pointer->asl_pointer))->p.Fulhes)((ASL*)(my_asl_pointer->asl_pointer),stk(l_W),n,0,0,stk(l_v));

  LhsVar(1) = Rhs + 1;

  return 0;
}

////////////////////////////////////
// Sparse version of the Jacobian //
////////////////////////////////////

int sci_ampl_eval_sp_g(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_x,       n_x,       l_x;
  int m_g,       n_g,       l_g;
  int m_J1,      n_J1;
  int n, nc, nz, nerror, i;
  char   * what        = NULL;
  double * tmp_dbl     = NULL;
  int    * colstarts   = NULL;
  struct my_asl * my_asl_pointer = NULL;
  cgrad  * cg = NULL, ** cgp = NULL;
  Jmp_buf err_jmp0;
  static SciSparse Jsp;

  if (Rhs!=2) 
    {
      sciprint("[g,spJac] = ampl_eval_sp_g(asl,x)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }
      
  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;
  nz = my_asl_pointer->asl_pointer->i.nzc_;

  CheckVector(2,n,1);

  GetRhsVar(2, "d", &m_x, &n_x, &l_x);

  nerror = -1;
  my_asl_pointer->asl_pointer->i.err_jmp1_ = &err_jmp0;

  m_g  = n;  n_g  = 1;
  CreateVar(Rhs+1, "d", &m_g,  &n_g,  &l_g);

  if (setjmp(err_jmp0.jb))
    {
      sciprint("ampl_evalg: trouble evaluating %s\n", what);
      return 0;
    }

  // Gradient of the objective function (dense)
  what = "g";
  (*((ASL*)(my_asl_pointer->asl_pointer))->p.Objgrd)((ASL*)(my_asl_pointer->asl_pointer),0,stk(l_x),stk(l_g),&nerror);

  // Jacobian of the constraints (sparse)

  Jsp.m    = n;
  Jsp.n    = nc;
  Jsp.nel  = nz;
  Jsp.mnel = (int *)MALLOC(n*sizeof(int));
  Jsp.icol = (int *)MALLOC(nz*sizeof(int));
  Jsp.R    = (double *)MALLOC(nz*sizeof(double));

  if (nc) 
    {
      what = "J";

      memset(Jsp.R,0,nz*sizeof(double));

      (*((ASL*)(my_asl_pointer->asl_pointer))->p.Jacval)((ASL*)(my_asl_pointer->asl_pointer),stk(l_x), Jsp.R, &nerror); // Jsp.R est de taille nz

      for(i=0;i<n;i++)
	{
	  Jsp.mnel[i] = my_asl_pointer->asl_pointer->i.A_colstarts_[i+1] - my_asl_pointer->asl_pointer->i.A_colstarts_[i];
	}

      cgp = my_asl_pointer->asl_pointer->i.Cgrad_;
      for(i=0;i<nc;i++)
	for(cg=*cgp++;cg;cg=cg->next)
	  {
	    Jsp.icol[cg->goff] = i + 1;
	  }
    }
  
#ifdef DEBUG
  DBGPRINTF("DEBUG: n = %d nc = %d nz = %d Index = %d\n", n, nc, nz, Index);
#endif

  m_J1 = Jsp.m; n_J1 = Jsp.n;
  CreateVarFromPtr(Rhs+2, "s", &m_J1, &n_J1, &Jsp);

  LhsVar(1) = Rhs + 1;
  LhsVar(2) = Rhs + 2;
  
  if (Jsp.mnel) FREE(Jsp.mnel);
  if (Jsp.icol) FREE(Jsp.icol);
  if (Jsp.R)    FREE(Jsp.R);

  return 0;
}

///////////////////////////////////
// Sparse version of the Hessian //
///////////////////////////////////

int sci_ampl_eval_sp_w(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_v,       n_v,       l_v;
  int m_W,       n_W;
  int n, nc, nz, nhnz;
  int nerror, i;
  char * what;
  struct my_asl * my_asl_pointer = NULL;
  Jmp_buf err_jmp0;
  static SciSparse Hsp;

  if ((Lhs!=1)||(Rhs!=2))
    {
      sciprint("spW = ampl_eval_sp_w(asl, v)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;
  nz = my_asl_pointer->asl_pointer->i.nzc_;

  CheckVector(2,nc,1);
  
  nerror = -1;
  my_asl_pointer->asl_pointer->i.err_jmp1_ = &err_jmp0;

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }

  GetRhsVar(2, "d", &m_v, &n_v, &l_v);
  
  if (setjmp(err_jmp0.jb))
    {
      sciprint("ampl_eval_sp_w: trouble evaluating %s\n", what);
      return 0;
    }

  nhnz = (*((ASL*)(my_asl_pointer->asl_pointer))->p.Sphset)((ASL*)(my_asl_pointer->asl_pointer), 0, 0, 0, nc > 0, 0);

  Hsp.m    = n;
  Hsp.n    = n;
  Hsp.nel  = nhnz;
  Hsp.mnel = (int *)MALLOC(n*sizeof(int));
  Hsp.icol = (int *)MALLOC(nhnz*sizeof(int));
  Hsp.R    = (double *)MALLOC(nhnz*sizeof(double));

  my_asl_pointer->asl_pointer->i.sputinfo_->uptri = 0; // We want all the Hessian (upper and lower triangles)

  what = "W";
  if (nc > 0) (*((ASL*)(my_asl_pointer->asl_pointer))->p.Sphes)((ASL*)(my_asl_pointer->asl_pointer),0,Hsp.R,0,0,stk(l_v));
  else        (*((ASL*)(my_asl_pointer->asl_pointer))->p.Sphes)((ASL*)(my_asl_pointer->asl_pointer),0,Hsp.R,0,0,0);

  for(i=0;i<n;i++) Hsp.mnel[i] = my_asl_pointer->asl_pointer->i.sputinfo_->hcolstarts[i+1] - my_asl_pointer->asl_pointer->i.sputinfo_->hcolstarts[i];
  // YC: tester sputinfo == NULL et hrownos == NULL
  for(i=0;i<nhnz;i++) Hsp.icol[i] = my_asl_pointer->asl_pointer->i.sputinfo_->hrownos[i]+1;

  m_W = n; n_W = n;
  CreateVarFromPtr(Rhs+1, "s", &Hsp.m, &Hsp.n, &Hsp);

  LhsVar(1) = Rhs + 1;

  if (Hsp.mnel) FREE(Hsp.mnel);
  if (Hsp.icol) FREE(Hsp.icol);
  if (Hsp.R)    FREE(Hsp.R);

  return 0;
}

////////////////////////////////////////////////////
// Returns the sparsity structure of the Jacobian //
////////////////////////////////////////////////////

int sci_ampl_eval_spst_g_rc(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_x,       n_x,       l_x;
  int m_irow,    n_irow,    l_irow;
  int m_jcol,    n_jcol,    l_jcol;
  int n, nc, nz, nerror, i, j, Index;
  char   * what        = NULL;
  double * tmp_dbl     = NULL;
  struct my_asl * my_asl_pointer = NULL;
  cgrad  * cg = NULL, ** cgp = NULL;
  Jmp_buf err_jmp0;

  if (Rhs!=2) 
    {
      sciprint("[irow,jcol] = ampl_eval_spst_g_rc(asl,x)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }
      
  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;
  nz = my_asl_pointer->asl_pointer->i.nzc_;

  CheckVector(2,n,1);

  GetRhsVar(2, "d", &m_x, &n_x, &l_x);

  nerror = -1;
  my_asl_pointer->asl_pointer->i.err_jmp1_ = &err_jmp0;

  if (setjmp(err_jmp0.jb))
    {
      sciprint("ampl_eval_spst_g_rc: trouble evaluating %s\n", what);
      return 0;
    }

  // Sparsity structure of the Jacobian

  m_irow = nz; n_irow = 1;
  m_jcol = nz; n_jcol = 1;
  
  CreateVar(3, "i", &m_irow, &n_irow, &l_irow);
  CreateVar(4, "i", &m_jcol, &n_jcol, &l_jcol);

  tmp_dbl = (double *)MALLOC(nz*sizeof(double));

  if (nc) 
    {
      what = "J";

      (*((ASL*)(my_asl_pointer->asl_pointer))->p.Jacval)((ASL*)(my_asl_pointer->asl_pointer),stk(l_x), tmp_dbl, &nerror);

      Index = 0;
      for(i=0;i<n;i++)
	{
	  for(j=0;j<my_asl_pointer->asl_pointer->i.A_colstarts_[i+1] - my_asl_pointer->asl_pointer->i.A_colstarts_[i];j++)
	    {
	      *istk(l_jcol + Index) = i + 1;
	      Index++;
	    }
	}

      cgp = my_asl_pointer->asl_pointer->i.Cgrad_;
      for(i=0;i<nc;i++)
	for(cg=*cgp++;cg;cg=cg->next)
	  {
	    *istk(l_irow + cg->goff) = i + 1;
	  }
    }
  
  LhsVar(1) = 3;
  LhsVar(2) = 4;
  
  if (tmp_dbl) FREE(tmp_dbl);

  return 0;
}

////////////////////////////////////////////////////////////////////////////////
// Returns the values corresponding to the sparsity structure of the Jacobian //
////////////////////////////////////////////////////////////////////////////////

int sci_ampl_eval_spst_g_val(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_x,       n_x,       l_x;
  int m_val,     n_val,     l_val;
  int n, nc, nz, nerror;
  char   * what        = NULL;
  double * tmp_dbl     = NULL;
  struct my_asl * my_asl_pointer = NULL;
  cgrad  * cg = NULL, ** cgp = NULL;
  Jmp_buf err_jmp0;

  if (Rhs!=2) 
    {
      sciprint("val = ampl_eval_spst_g_val(asl,x)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }
      
  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;
  nz = my_asl_pointer->asl_pointer->i.nzc_;

  CheckVector(2,n,1);

  GetRhsVar(2, "d", &m_x, &n_x, &l_x);

  nerror = -1;
  my_asl_pointer->asl_pointer->i.err_jmp1_ = &err_jmp0;

  if (setjmp(err_jmp0.jb))
    {
      sciprint("ampl_eval_spst_g_val: trouble evaluating %s\n", what);
      return 0;
    }

  // Sparsity structure of the Jacobian

  m_val = nz; n_val = 1;
  
  CreateVar(3, "d", &m_val, &n_val, &l_val);
  memset(stk(l_val),0,nz*sizeof(double));

  if (nc) 
    {
      what = "J";

      (*((ASL*)(my_asl_pointer->asl_pointer))->p.Jacval)((ASL*)(my_asl_pointer->asl_pointer),stk(l_x), stk(l_val), &nerror);
    }

  LhsVar(1) = 3;
  
  return 0;
}

///////////////////////////////////////////////////
// Returns the sparsity structure of the Hessian //
///////////////////////////////////////////////////

int sci_ampl_eval_spst_w_rc(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_v,       n_v,       l_v;
  int m_irow,    n_irow,    l_irow;
  int m_jcol,    n_jcol,    l_jcol;
  int n, nc, nz, nhnz;
  int nerror, i, j, Index;
  char   * what        = NULL;
  double * tmp_dbl     = NULL;
  struct my_asl * my_asl_pointer = NULL;
  Jmp_buf err_jmp0;

  if ((Lhs!=1)||(Rhs!=2))
    {
      sciprint("[irow,jcol] = ampl_eval_spst_w_rc(asl, v)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;
  nz = my_asl_pointer->asl_pointer->i.nzc_;

  CheckVector(2,nc,1);
  
  nerror = -1;
  my_asl_pointer->asl_pointer->i.err_jmp1_ = &err_jmp0;

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }

  GetRhsVar(2, "d", &m_v, &n_v, &l_v);
  
  if (setjmp(err_jmp0.jb))
    {
      sciprint("ampl_eval_spst_w_rc: trouble evaluating %s\n", what);
      return 0;
    }

  // Sparsity structure of the Jacobian

  m_irow = nz; n_irow = 1;
  m_jcol = nz; n_jcol = 1;
  
  CreateVar(3, "i", &m_irow, &n_irow, &l_irow);
  CreateVar(4, "i", &m_jcol, &n_jcol, &l_jcol);

  tmp_dbl = (double *)MALLOC(nhnz*sizeof(double));

  nhnz = (*((ASL*)(my_asl_pointer->asl_pointer))->p.Sphset)((ASL*)(my_asl_pointer->asl_pointer), 0, 0, 0, nc > 0, 0);

  my_asl_pointer->asl_pointer->i.sputinfo_->uptri = 0; // We want all the Hessian (upper and lower triangles)

  what = "W";
  if (nc > 0) (*((ASL*)(my_asl_pointer->asl_pointer))->p.Sphes)((ASL*)(my_asl_pointer->asl_pointer),0,tmp_dbl,0,0,stk(l_v));
  else        (*((ASL*)(my_asl_pointer->asl_pointer))->p.Sphes)((ASL*)(my_asl_pointer->asl_pointer),0,tmp_dbl,0,0,0);

  Index = 0;
  for(i=0;i<n;i++)
    {
      for(j=0;j<my_asl_pointer->asl_pointer->i.sputinfo_->hcolstarts[i+1] - my_asl_pointer->asl_pointer->i.sputinfo_->hcolstarts[i];j++)
	{
	  *istk(l_jcol + Index) = i + 1;
	  Index++;
	}
    }

  for(i=0;i<nhnz;i++) *istk(l_irow + i) = my_asl_pointer->asl_pointer->i.sputinfo_->hrownos[i] + 1;

  LhsVar(1) = 3;
  LhsVar(2) = 4;
  
  if (tmp_dbl) FREE(tmp_dbl);

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Returns the values corresponding to the sparsity structure of the Hessian //
///////////////////////////////////////////////////////////////////////////////

int sci_ampl_eval_spst_w_val(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_v,       n_v,       l_v;
  int m_val,     n_val,     l_val;
  int n, nc, nz, nhnz;
  int nerror;
  char   * what        = NULL;
  struct my_asl * my_asl_pointer = NULL;
  Jmp_buf err_jmp0;

  if ((Lhs!=1)||(Rhs!=2))
    {
      sciprint("val = ampl_eval_spst_w_val(asl, v)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;
  nz = my_asl_pointer->asl_pointer->i.nzc_;

  CheckVector(2,nc,1);
  
  nerror = -1;
  my_asl_pointer->asl_pointer->i.err_jmp1_ = &err_jmp0;

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }

  GetRhsVar(2, "d", &m_v, &n_v, &l_v);
  
  if (setjmp(err_jmp0.jb))
    {
      sciprint("ampl_eval_spst_w_val: trouble evaluating %s\n", what);
      return 0;
    }

  // Sparsity structure of the Jacobian

  m_val = nhnz; n_val = 1;
  
  CreateVar(3, "d", &m_val, &n_val, &l_val);

  nhnz = (*((ASL*)(my_asl_pointer->asl_pointer))->p.Sphset)((ASL*)(my_asl_pointer->asl_pointer), 0, 0, 0, nc > 0, 0);

  my_asl_pointer->asl_pointer->i.sputinfo_->uptri = 0; // We want all the Hessian (upper and lower triangles)

  what = "W";
  if (nc > 0) (*((ASL*)(my_asl_pointer->asl_pointer))->p.Sphes)((ASL*)(my_asl_pointer->asl_pointer),0,stk(l_val),0,0,stk(l_v));
  else        (*((ASL*)(my_asl_pointer->asl_pointer))->p.Sphes)((ASL*)(my_asl_pointer->asl_pointer),0,stk(l_val),0,0,0);

  LhsVar(1) = 3;

  return 0;
}

///////////////////////
// Utility functions //
///////////////////////

int sci_ampl_write_sol(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_string,  n_string,  l_string;
  int m_x,       n_x,       l_x;
  int m_v,       n_v,       l_v;
  struct my_asl * my_asl_pointer = NULL;
  int n, nc;

  /* ampl_write_sol(asl, 'solution message', x, v): x = primal, v = dual */

  if ((GetType(2)!=sci_strings)||(Rhs!=4))
    {
      sciprint("ampl_write_sol(asl, 'solution message', x, v)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  n  = my_asl_pointer->asl_pointer->i.n_var_;
  nc = my_asl_pointer->asl_pointer->i.n_con_;

  CheckVector(3,n,1);
  CheckVector(4,nc,1);

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }

  GetRhsVar(2, "c", &m_string, &n_string, &l_string);
  GetRhsVar(3, "d", &m_x, &n_x, &l_x);
  GetRhsVar(4, "d", &m_v, &n_v, &l_v);

  write_sol_ASL((ASL*)(my_asl_pointer->asl_pointer),cstk(l_string),stk(l_x),stk(l_v),0);
  
  return 0;
}

//////////////////////////////////////////////////
// Get informations related to the AMPL problem //
//////////////////////////////////////////////////

int sci_ampl_get_size(char * fname)
{
  int m_extra  = 22, n_extra  = 1, l_extra; 
  int m_nbv    = 1,  n_nbv    = 1, l_list_nbv,    l_nbv;
  int m_niv    = 1,  n_niv    = 1, l_list_niv,    l_niv;
  int m_nlc    = 1,  n_nlc    = 1, l_list_nlc,    l_nlc;
  int m_n_eqn  = 1,  n_n_eqn  = 1, l_list_n_eqn,  l_n_eqn;
  int m_n_cc   = 1,  n_n_cc   = 1, l_list_n_cc,   l_n_cc;
  int m_nlcc   = 1,  n_nlcc   = 1, l_list_nlcc,   l_nlcc;
  int m_nlnc   = 1,  n_nlnc   = 1, l_list_nlnc,   l_nlnc;
  int m_nlo    = 1,  n_nlo    = 1, l_list_nlo,    l_nlo;
  int m_nlvb   = 1,  n_nlvb   = 1, l_list_nlvb,   l_nlvb;
  int m_nlvc   = 1,  n_nlvc   = 1, l_list_nlvc,   l_nlvc;
  int m_nlvo   = 1,  n_nlvo   = 1, l_list_nlvo,   l_nlvo;
  int m_nlvbi  = 1,  n_nlvbi  = 1, l_list_nlvbi,  l_nlvbi;
  int m_nlvci  = 1,  n_nlvci  = 1, l_list_nlvci,  l_nlvci;
  int m_nlvoi  = 1,  n_nlvoi  = 1, l_list_nlvoi,  l_nlvoi;
  int m_nwv    = 1,  n_nwv    = 1, l_list_nwv,    l_nwv;
  int m_nzc    = 1,  n_nzc    = 1, l_list_nzc,    l_nzc;
  int m_nzo    = 1,  n_nzo    = 1, l_list_nzo,    l_nzo;
  int m_n_var  = 1,  n_n_var  = 1, l_list_n_var,  l_n_var;
  int m_n_con  = 1,  n_n_con  = 1, l_list_n_con,  l_n_con;
  int m_n_obj  = 1,  n_n_obj  = 1, l_list_n_obj,  l_n_obj;
  int m_n_lcon = 1,  n_n_lcon = 1, l_list_n_lcon, l_n_lcon;
  int m_pointer,     n_pointer,    l_pointer;
  int m_list_labels,  n_list_labels;

  static const char * ListLabels [] = {"plist",
				       "nbv", /* no. of linear binary variables */
				       "niv", /* no. of linear integer variables */
				       "nlc", /* total no. of nonlinear constraints */
				       "n_eqn", /* number of equality constraints or -1 if unknown (ampl prior to 19970627) */
				       "n_cc", /* total complementarity conditions */
				       "nlcc", /* nonlinear complementarity conditions */
				       "nlnc", /* no. of nonlinear network constraints */
				       "nlo", /* no. of nonlinear objectives */
				       "nlvb", /* no. of nonlinear variables in both constraints and objectives */
				       "nlvc", /* no. of nonlinear variables in constraints */
				       "nlvo", /* no. of nonlinear variables in objectives nlvc_ and nlvo_ include nlvb_ */
				       "nlvbi", /* integer nonlinear variables in both constraints and objectives */
				       "nlvci", /* integer nonlinear vars just in constraints */
				       "nlvoi", /* integer nonlinear vars just in objectives */
				       "nwv", /* no. of (linear) network variables (arcs) */
				       "nzc", /* no. of nonzeros in constraints' Jacobian */
				       "nzo", /* no. of nonzeros in all objective gradients */
				       "n_var", /* total no. of variables */
				       "n_con", /* total no. of constraints */
				       "n_obj", /* total no. of objectives */
				       "n_lcon"}; /* no. of logical constraints */
  struct my_asl * my_asl_pointer = NULL;

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }

  if ((Lhs!=1)||(Rhs!=1))
    {
      sciprint("info = ampl_get_size(asl)\n");
      return 0;
    }

  CreateVar(Rhs + 1,  "i", &m_nbv,    &n_nbv,    &l_nbv);
  CreateVar(Rhs + 2,  "i", &m_niv,    &n_niv,    &l_niv);
  CreateVar(Rhs + 3,  "i", &m_nlc,    &n_nlc,    &l_nlc);
  CreateVar(Rhs + 4,  "i", &m_n_eqn,  &n_n_eqn,  &l_n_eqn);
  CreateVar(Rhs + 5,  "i", &m_n_cc,   &n_n_cc,   &l_n_cc);
  CreateVar(Rhs + 6,  "i", &m_nlcc,   &n_nlcc,   &l_nlcc);
  CreateVar(Rhs + 7,  "i", &m_nlnc,   &n_nlnc,   &l_nlnc);
  CreateVar(Rhs + 8,  "i", &m_nlo,    &n_nlo,    &l_nlo);
  CreateVar(Rhs + 9,  "i", &m_nlvb,   &n_nlvb,   &l_nlvb);
  CreateVar(Rhs + 10, "i", &m_nlvc,   &n_nlvc,   &l_nlvc);
  CreateVar(Rhs + 11, "i", &m_nlvo,   &n_nlvo,   &l_nlvo);
  CreateVar(Rhs + 12, "i", &m_nlvbi,  &n_nlvbi,  &l_nlvbi);
  CreateVar(Rhs + 13, "i", &m_nlvci,  &n_nlvci,  &l_nlvci);
  CreateVar(Rhs + 14, "i", &m_nlvoi,  &n_nlvoi,  &l_nlvoi);
  CreateVar(Rhs + 15, "i", &m_nwv,    &n_nwv,    &l_nwv);
  CreateVar(Rhs + 16, "i", &m_nzc,    &n_nzc,    &l_nzc);
  CreateVar(Rhs + 17, "i", &m_nzo,    &n_nzo,    &l_nzo);
  CreateVar(Rhs + 18, "i", &m_n_var,  &n_n_var,  &l_n_var);
  CreateVar(Rhs + 19, "i", &m_n_con,  &n_n_con,  &l_n_con);
  CreateVar(Rhs + 20, "i", &m_n_obj,  &n_n_obj,  &l_n_obj);
  CreateVar(Rhs + 21, "i", &m_n_lcon, &n_n_lcon, &l_n_lcon);

  *istk(l_nbv)    = my_asl_pointer->asl_pointer->i.nbv_;
  *istk(l_niv)    = my_asl_pointer->asl_pointer->i.niv_;
  *istk(l_nlc)    = my_asl_pointer->asl_pointer->i.nlc_;
  *istk(l_n_eqn)  = my_asl_pointer->asl_pointer->i.n_eqn_;
  *istk(l_n_cc)   = my_asl_pointer->asl_pointer->i.n_cc_;
  *istk(l_nlcc)   = my_asl_pointer->asl_pointer->i.nlcc_;
  *istk(l_nlnc)   = my_asl_pointer->asl_pointer->i.nlnc_;
  *istk(l_nlo)    = my_asl_pointer->asl_pointer->i.nlo_;
  *istk(l_nlvb)   = my_asl_pointer->asl_pointer->i.nlvb_;
  *istk(l_nlvc)   = my_asl_pointer->asl_pointer->i.nlvc_;
  *istk(l_nlvo)   = my_asl_pointer->asl_pointer->i.nlvo_;
  *istk(l_nlvbi)  = my_asl_pointer->asl_pointer->i.nlvbi_;
  *istk(l_nlvci)  = my_asl_pointer->asl_pointer->i.nlvci_;
  *istk(l_nlvoi)  = my_asl_pointer->asl_pointer->i.nlvoi_;
  *istk(l_nwv)    = my_asl_pointer->asl_pointer->i.nwv_;
  *istk(l_nzc)    = my_asl_pointer->asl_pointer->i.nzc_;
  *istk(l_nzo)    = my_asl_pointer->asl_pointer->i.nzo_;
  *istk(l_n_var)  = my_asl_pointer->asl_pointer->i.n_var_;
  *istk(l_n_con)  = my_asl_pointer->asl_pointer->i.n_con_;
  *istk(l_n_obj)  = my_asl_pointer->asl_pointer->i.n_obj_;
  *istk(l_n_lcon) = my_asl_pointer->asl_pointer->i.n_lcon_;

  // Create the 'extra' structure of type plist
  CreateVar(Rhs + 22, "m", &m_extra, &n_extra, &l_extra);

  m_list_labels = 22; n_list_labels = 1;
  CreateListVarFromPtr(Rhs + 22, 1, "S", &m_list_labels, &n_list_labels, ListLabels);
  CreateListVarFrom(Rhs + 22, 2,  "i", &m_nbv,    &n_nbv,    &l_list_nbv,    &l_nbv);
  CreateListVarFrom(Rhs + 22, 3,  "i", &m_niv,    &n_niv,    &l_list_niv,    &l_niv);
  CreateListVarFrom(Rhs + 22, 4,  "i", &m_nlc,    &n_nlc,    &l_list_nlc,    &l_nlc);
  CreateListVarFrom(Rhs + 22, 5,  "i", &m_n_eqn,  &n_n_eqn,  &l_list_n_eqn,  &l_n_eqn);
  CreateListVarFrom(Rhs + 22, 6,  "i", &m_n_cc,   &n_n_cc,   &l_list_n_cc,   &l_n_cc);
  CreateListVarFrom(Rhs + 22, 7,  "i", &m_nlcc,   &n_nlcc,   &l_list_nlcc,   &l_nlcc);
  CreateListVarFrom(Rhs + 22, 8,  "i", &m_nlnc,   &n_nlnc,   &l_list_nlnc,   &l_nlnc);
  CreateListVarFrom(Rhs + 22, 9,  "i", &m_nlo,    &n_nlo,    &l_list_nlo,    &l_nlo);
  CreateListVarFrom(Rhs + 22, 10, "i", &m_nlvb,   &n_nlvb,   &l_list_nlvb,   &l_nlvb);
  CreateListVarFrom(Rhs + 22, 11, "i", &m_nlvc,   &n_nlvc,   &l_list_nlvc,   &l_nlvc);
  CreateListVarFrom(Rhs + 22, 12, "i", &m_nlvo,   &n_nlvo,   &l_list_nlvo,   &l_nlvo);
  CreateListVarFrom(Rhs + 22, 13, "i", &m_nlvbi,  &n_nlvbi,  &l_list_nlvbi,  &l_nlvbi);
  CreateListVarFrom(Rhs + 22, 14, "i", &m_nlvci,  &n_nlvci,  &l_list_nlvci,  &l_nlvci);
  CreateListVarFrom(Rhs + 22, 15, "i", &m_nlvoi,  &n_nlvoi,  &l_list_nlvoi,  &l_nlvoi);
  CreateListVarFrom(Rhs + 22, 16, "i", &m_nwv,    &n_nwv,    &l_list_nwv,    &l_nwv);
  CreateListVarFrom(Rhs + 22, 17, "i", &m_nzc,    &n_nzc,    &l_list_nzc,    &l_nzc);
  CreateListVarFrom(Rhs + 22, 18, "i", &m_nzo,    &n_nzo,    &l_list_nzo,    &l_nzo);
  CreateListVarFrom(Rhs + 22, 19, "i", &m_n_var,  &n_n_var,  &l_list_n_var,  &l_n_var);
  CreateListVarFrom(Rhs + 22, 20, "i", &m_n_con,  &n_n_con,  &l_list_n_con,  &l_n_con);
  CreateListVarFrom(Rhs + 22, 21, "i", &m_n_obj,  &n_n_obj,  &l_list_n_obj,  &l_n_obj);
  CreateListVarFrom(Rhs + 22, 22, "i", &m_n_lcon, &n_n_lcon, &l_list_n_lcon, &l_n_lcon);

  LhsVar(1) = Rhs + 22;

  return 0;
}

/////////////////////////////////////
// Get the complementarity indexes //
/////////////////////////////////////

int sci_ampl_get_compl(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_cvar,    n_cvar,    l_cvar;
  int nc, i;
  struct my_asl * my_asl_pointer = NULL;

  if ((Lhs!=1)||(Rhs!=1))
    {
      sciprint("cvar = ampl_get_compl(asl)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }

  nc = my_asl_pointer->asl_pointer->i.n_con_;

  if (!my_asl_pointer->asl_pointer->i.n_cc_)
    {
      m_cvar = 1; n_cvar = 1;
      CreateVar(Rhs+1, "d", &m_cvar, &n_cvar, &l_cvar);

      *stk(l_cvar) = -1;

      LhsVar(1) = Rhs + 1;

      return 0;
    }
      
  /* cvar[i] > 0 means constraint i complements variable cvar[i] - 1 */
  
  m_cvar = nc; n_cvar = 1;
  CreateVar(Rhs+1, "d", &m_cvar, &n_cvar, &l_cvar);

  for(i=0;i<nc;i++) *stk(l_cvar + i) = my_asl_pointer->asl_pointer->i.cvar_[i];

  LhsVar(1) = Rhs + 1;

  return 0;
}

///////////////////////////////
// Get the type of variables //
///////////////////////////////

int sci_ampl_get_type(char * fname)
{
  int m_pointer, n_pointer, l_pointer;
  int m_type,    n_type,    l_type;
  int n, i, i_start, i_end, tmp_int;
  struct my_asl * my_asl_pointer = NULL;

  if ((Lhs!=1)||(Rhs!=1))
    {
      sciprint("var_type = ampl_get_type(asl)\n");
      return 0;
    }

  GetRhsVar(1, "p", &m_pointer, &n_pointer, &l_pointer);

  my_asl_pointer = (struct my_asl *)((unsigned long)*stk(l_pointer));

  if (!my_asl_pointer->asl_pointer->i.filename_)
    {
      sciprint("ampl_init(filename) has not been called\n");
      return 0;
    }
      
  n  = my_asl_pointer->asl_pointer->i.n_var_;

  m_type = n; n_type = 1;
  CreateVar(Rhs + 1, "c", &m_type, &n_type, &l_type);

  tmp_int = (my_asl_pointer->asl_pointer->i.nlvc_<my_asl_pointer->asl_pointer->i.nlvo_)?my_asl_pointer->asl_pointer->i.nlvo_:my_asl_pointer->asl_pointer->i.nlvc_;

  i_start = 0;
  i_end   = i_start + tmp_int;
  
  for(i=i_start; i<i_end;i++) *cstk(l_type + i) = 'n'; // n: nonlinear variable

  i_start = i_end;
  i_end   = i_start + my_asl_pointer->asl_pointer->i.nwv_;

  for(i=i_start; i<i_end;i++) *cstk(l_type + i) = 'a'; // a: linear arcs

  i_start = i_end;
  i_end   = i_start + my_asl_pointer->asl_pointer->i.n_var_ - (tmp_int + 
							       my_asl_pointer->asl_pointer->i.niv_ + 
							       my_asl_pointer->asl_pointer->i.nbv_ + 
							       my_asl_pointer->asl_pointer->i.nwv_);

  for(i=i_start; i<i_end;i++) *cstk(l_type + i) = 'o'; // o: other linear

  i_start = i_end;
  i_end   = i_start + my_asl_pointer->asl_pointer->i.nbv_;

  for(i=i_start; i<i_end;i++) *cstk(l_type+i) = 'b'; // b: binary variable

  i_start = i_end;
  i_end   = i_start + my_asl_pointer->asl_pointer->i.niv_;

  for(i=i_start; i<i_end;i++) *cstk(l_type+i) = 'i'; // i: integer variable

#ifdef DEBUG
  DBGPRINTF("DEBUG: i_end = %d n = %d\n", i_end, n);
#endif

  LhsVar(1) = Rhs + 1;

  return 0;
}
